#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>

#include "sun.h"

Log *Log::open(const char *filename, int loglevel, bool append)
{
  FILE *file = fopen(filename, append ? "a" : "w+");
  if (!file)
    return NULL;
    
  return new Log(file, loglevel);
}

Log *Log::openEmpty()
{
  return new Log(NULL, 99);
}

Log::Log(FILE *file, int loglevel)
{
  this->file = file;
  this->loglevel = loglevel;
  this->logFlush = false;
}

void Log::close()
{  
  if (file) {
    fclose(file);
    file = NULL;
  }
}

Log::~Log()
{
  if (file)
    close();
}

void Log::write(int level, const char *format, ...)
{
  if (!file)
    return;
  if ((level>0) && (level>loglevel))
    return;

  char buffer[4096];
  va_list ap;
  va_start (ap, format);
  if (vsnprintf (buffer, sizeof(buffer), format, ap) > (int)sizeof(buffer))
    panic("Buffer overflow in log()");

  char indent[100];
  if (level == 0) 
    strcpy(indent, "");
  else if (level == 1)
    strcpy(indent, "=== ");
  else if (level == 2)
    strcpy(indent, "  ");
  else if (level == 3)
    strcpy(indent, "    ");
  else if (level == 4)
    strcpy(indent, "      ");
  else
    panic("Level=%d", level);

  char prefix[200];

  time_t ti = time(0);
  struct tm ttm;
  gmtime_r(&ti, &ttm);
  sprintf(prefix, "%02d/%02d/%04d %02d:%02d:%02d %s", 
    1+ttm.tm_mon, ttm.tm_mday, 1900+ttm.tm_year, ttm.tm_hour, ttm.tm_min, ttm.tm_sec,
    indent
  );

  char *ptr = buffer;
  while (true) {
    char *lf = strchr(ptr, '\n');
    if (!lf) {
      fprintf(file, "%s%s\n", prefix, ptr);
      break;
    }

    *lf = 0;
    fprintf(file, "%s%s\n", prefix, ptr);
    ptr = lf+1;
    if (!*ptr)
      break;
  }

  if (logFlush)
    fflush(file);
    
  va_end (ap);
}

void Log::dump(int level, unsigned char *payload, int len)
{
  if (!file)
    return;
  if ((level>0) && (level>loglevel))
    return;
    
  char outbuf[100];
  int i, off = 0;
  while (off < len) {
    sprintf(outbuf, "%04X   ", off);
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        sprintf(&outbuf[7+3*i], "%02X ", payload[i+off]);
      else
        strcat(outbuf, "   ");
    }
    
    strcat(outbuf, "   ");
    outbuf[58] = 0;
    
    for (i=0; i<16; i++) {
      if ((i+off) < len) {
        outbuf[58+i] = isprint(payload[i+off]) ? payload[i+off] : '.';
        outbuf[59+i] = 0;
      }
    }
    
    off += 16;
    write(level, "%s", outbuf);
  }
}

