#include <stdlib.h>
#include <string.h>

#include "sun.h"
#include "tools.h"

int main(int argc, char *argv[]) 
{
  srandom(time(0));

  if (argc != 2) 
    panic("Syntax: %s <parameterFile>", argv[0]);
    
  ParameterFile *params = new ParameterFile(argv[1]);
  if (!params->selectSection("settings"))
    panic("Section [settings] does not exist in %s", argv[1]);

  /* Log file */
    
  char val[200];
  Log *log;
    
  if (params->existsSetting("logFile")) {
    params->getAsString("logFile", val, sizeof(val));

    bool doAppend = true;
    if (params->existsSetting("append"))
      doAppend = params->getAsBoolean("append");

    int loglevel = 1;
    if (params->existsSetting("logLevel"))
      loglevel = params->getAsInt("logLevel");
    
    log = Log::open(val, loglevel, doAppend);
    if (!log)
      panic("Cannot open log file '%s'", val);
      
    if (params->existsSetting("logFlush"))
      log->setFlush(params->getAsBoolean("logFlush"));
  } else {
    log = Log::openEmpty();
  }
  
  /* Components */
  
  Selector *selector = new Selector();

  Database *db = NULL;
  if (params->existsSetting("database")) {
    params->getAsString("database", val, sizeof(val));
    if (!strcmp(val, "file"))
      db = new FileDatabase(params, selector, log);
    else if (!strcmp(val, "dynamic"))
      db = new DynamicDatabase(params, selector, log);
    else if (!strcmp(val, "mysql"))
      db = new MysqlDatabase(params, selector, log);
  }
  
  if (!db)
    panic("Please specify either database=file or database=mysql in %s!", argv[1]);

  Sun *sun = new Sun(params, selector, log, db);
  sun->init();
  
  WebFrontend *frontend = NULL;
  if (params->existsSetting("webInterface") && params->getAsBoolean("webInterface")) {
    frontend = new WebFrontend(params, selector, log, sun);
    frontend->init();
  }
  
  if (params->existsSetting("daemon") && params->getAsBoolean("daemon"))
    becomeDaemon();

  /* Main loop */
  
  selector->run();
  return 0;
}
