#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <openssl/sha.h>

#include "sun.h"
#include "tools.h"

#define append(x...) do { char xbuf[1000]; snprintf(xbuf, sizeof(xbuf), x); xbuf[999]=0; int p=0; while (xbuf[p] && (connection[index].len<sizeof(connection[index].buf))) connection[index].buf[connection[index].len++] = xbuf[p++]; } while (0)

WebFrontend::WebFrontend(ParameterFile *params, Selector *selector, Log *log, Sun *sun)
{
  this->selector = selector;
  this->httpPort = params->existsSetting("httpPort") ? params->getAsInt("httpPort") : DEFAULT_HTTP_PORT;
  this->hostname[0] = 0;
  this->httpSocket = -1;
  this->sun = sun;
  this->log = log;
  
  for (int i=0; i<MAX_CONNECTIONS; i++)
    connection[i].active = false;
    
  char val1[200], val2[200];
  if (params->existsSetting("adminUsername"))
    params->getAsString("adminUsername", val1, sizeof(val1));
  else
    strcpy(val1, "admin");
    
  if (params->existsSetting("adminPassword"))
    params->getAsString("adminPassword", val2, sizeof(val2));
  else
    strcpy(val2, "admin");
    
  snprintf(adminAuth, sizeof(adminAuth), "%s:%s", val1, val2);
}

void WebFrontend::init()
{
  /* Find out the local hostname by one of several methods. This is important
     because the host name will be displayed in the status report. */

  if (!getenv("HOSTNAME")) {
    FILE *infile = fopen("/proc/sys/kernel/hostname", "r");
    if (infile) {
      fread(hostname, 1, sizeof(hostname), infile);
      fclose(infile);
    } else {
      strcpy(hostname, "localhost");
    }
  } else {
    strcpy(hostname, getenv("HOSTNAME"));
  }

  strtok(hostname, "\r\n");
  
  httpSocket = openServerSocket(httpPort, SOCK_STREAM);
  if (httpSocket < 0)
    panic("Cannot bind to socket %d/tcp (%s)", httpPort, strerror(errno));
    
  if (listen(httpSocket, 1)<0)
    panic("listen() failed\n");

  int on = 1;
  setsockopt(httpSocket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

  selector->setDescriptorCallback(httpSocket, true, false, false, this);
}

bool base64decode(const char *input, char *output, int maxOutputLen)
{
  const char *alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  int bits = 0;
  int haveBits = 0;
  int outputLen = 0;
  for (int i=0; input[i]; i++) {
    char *x = strchr(alpha, input[i]);
    if (!x) 
      break;
    
    int y = x-alpha;
    bits = (bits<<6)|y;
    haveBits += 6;
    if (haveBits >= 8) {
      if (outputLen < (maxOutputLen-1))
        output[outputLen++] = (bits>>(haveBits-8))&0xFF;

      haveBits -= 8;
      bits &= (1<<haveBits)-1;
    }
  }
  
  output[outputLen] = 0;
  return true;
}

void WebFrontend::authenticationFailed(int index, const char *authDomain)
{
  append("HTTP/1.1 401 UNAUTHORIZED\r\nServer: SatelliteLab\r\nWWW-Authenticate: Basic realm=\"%s\"\r\nContent-Type: text/html\r\nConnection: close\r\n\r\n", authDomain);
  append("<html><head><title>Authorization required</title></head><body><h3>Authorization required</h3></body></html>\n");
}

void WebFrontend::maybeHandleHttpRequest(int index)
{
  char buf1[256];

  /* Look for the CRLF-CRLF that indicates the end of the HTTP request */

  bool identifiedCommand = false;
  char *pattern = "\r\n\r\n";
  int ppos = 0;
  for (int i=0; (i<connection[index].len) && !identifiedCommand; i++) {
    if (connection[index].buf[i] == pattern[ppos])
      ppos ++;
    else 
      ppos = 0;
 
    if (ppos >= strlen(pattern))
      identifiedCommand = true;
  }
  
  if (!identifiedCommand)
    return;

  /* Find the authorization line, if one is present, and extract the password */

  char *firstline = strtok(connection[index].buf, "\r\n");
  char *item = firstline;
  char *authLine = NULL;
  if (item)
    item = strtok(NULL, "\r\n");
    
  while (item && item[0]) {
    if (!strncmp(item, "Authorization: Basic ", 21))
      authLine = item;
      
    item = strtok(NULL, "\r\n");
  }
  
  char auth[200];
  auth[0] = 0;
  if (authLine) 
    base64decode(&authLine[21], auth, sizeof(auth));
  bool isAdmin = !strcmp(auth, adminAuth);
  
  /* Parse the HTTP request */
  
  char *cmd = strtok(firstline, " ");
  char *curl = cmd ? strtok(NULL, " ") : NULL;
  if (curl && !strcmp(cmd, "GET")) {
    int now = time(0);
    char url[200];
    strncpy(url, curl, sizeof(url));
    
    connection[index].buf[0] = 0;
    connection[index].len = 0;

    /* Satellite reassociation */

    if (!strncmp(url, "/reassociate/", 13)) {
      char *owner = strtok(&url[13], "/");
      char *identifier = owner ? strtok(NULL, "/?\r\n") : NULL;
      if (owner && !owner[0])
        owner = NULL;
      if (identifier && !identifier[0])
        identifier = NULL;
      
      if (owner && identifier) {
        if (isAdmin || sun->isValidAuthentication(owner, auth)) {
          append("HTTP/1.1 200 OK\r\nServer: SatelliteLab\r\nContent-Type: text/html\r\nConnection: close\r\nRefresh: 0; URL=/experiments/%s/%s/\r\n\r\n", owner, identifier);
          sun->reassociateSatellites(owner, identifier);
        } else {
          append("HTTP/1.1 401 UNAUTHORIZED\r\nServer: SatelliteLab\r\nConnection: close\r\n\r\nNot authorized to reassociate satellites in %s:%s\n", owner, identifier);
        }
      } else {
        append("HTTP/1.1 404 NOT FOUND\r\nServer: SatelliteLab\r\nConnection: close\r\n\r\nNot found\n");
      }
    } else if (!strcmp(url, "/") || !strncmp(url, "/experiments/", 13)) {

      /* Overview page */

      char *ownerFilter = NULL;
      char *identifierFilter = NULL;
      
      if (!strncmp(url, "/experiments/", 13)) {
        ownerFilter = strtok(&url[13], "/");
        identifierFilter = ownerFilter ? strtok(NULL, "/?\r\n") : NULL;
        if (ownerFilter && !ownerFilter[0])
          ownerFilter = NULL;
        if (identifierFilter && !identifierFilter[0])
          identifierFilter = NULL;
      }
      
      /* Only the administrator is allowed to see everything; the others can just see their own experiments */
      
      if (isAdmin || (ownerFilter && sun->isValidAuthentication(ownerFilter, auth))) {
        append("HTTP/1.1 200 OK\r\nServer: SatelliteLab\r\nContent-Type: text/html\r\nConnection: close\r\n\r\n");
        append("<html><head><title>SatelliteLab testbed</title></head>\n<body>\n");
        
        /* List of experiments with some high-level stats */
        
        append("<h3>Experiments</h3>\n");
        append("  <table border=\"1\" cellpadding=\"3\" cellspacing=\"0\"><tr><td><b>Identifier</b></td><td><b>Status</b></td><td><b>Start</b></td><td><b>Time left</b></td><td><b>Planets</b></td><td><b>Satellites</b></td></tr>\n");
        for (int i=0; i<sun->numExperimentRecords(); i++) {
          struct Sun::experimentStruct exp;
          if (sun->statExperimentRecord(i, &exp)) {
            if ((!ownerFilter || !strcmp(exp.owner, ownerFilter)) && (!identifierFilter || !strcmp(exp.identifier, identifierFilter))) {
              append("<tr><td>");
              if (!ownerFilter)
                append("<a href=\"/experiments/%s\">%s</a>:<a href=\"/experiments/%s/%s\">%s</a>", exp.owner, exp.owner, exp.owner, exp.identifier, exp.identifier);
              else 
                append("<a href=\"/experiments/%s/%s\">%s</a>", exp.owner, exp.identifier, exp.identifier);
              append("</td>");
              if (exp.startTime>now)
                append("<td bgcolor=\"#FFEE00\" align=\"center\">INIT</td>");
              else if (exp.endTime<now)
                append("<td bgcolor=\"#0044FF\" align=\"center\">DONE</td>");
              else 
                append("<td bgcolor=\"#00B000\" align=\"center\">RUN</td>");

              time_t ti = exp.startTime;
              append("<td>%s</td>", ctime(&ti));
              int runtime = (exp.startTime>now) ? (exp.endTime-exp.startTime) : (exp.endTime-now);
              if (exp.endTime > now) 
                append("<td align=\"right\">%d:%02d:%02d</td>", runtime/3600, (runtime/60)%60, runtime%60);
              else
                append("<td align=\"right\">0:00:00</td>");

              int numSatellitesActive = 0;
              struct Sun::satelliteStruct sat;
              for (int j=0; j<sun->numSatelliteRecords(); j++) {
                sun->statSatelliteRecord(j, &sat);
                if (!strcmp(sat.expOwner, exp.owner) && !strcmp(sat.expIdentifier, exp.identifier) && (sat.lastStatusReport >= (now-Sun::SATELLITE_TIMEOUT_SEC)))
                  numSatellitesActive ++;
              }

              int numPlanetsActive = 0;
              struct Sun::planetStruct planet;
              for (int j=0; j<sun->numPlanetRecords(); j++) {
                sun->statPlanetRecord(j, &planet);
                if (!strcmp(planet.expOwner, exp.owner) && !strcmp(planet.expIdentifier, exp.identifier) && (planet.lastStatusReport >= (now-Sun::PLANET_TIMEOUT_SEC)))
                  numPlanetsActive ++;
              }

              const char *plcolor = "#FFFFFF";
              if ((exp.startTime>now) || (exp.endTime<now))
                plcolor = "#FFFFFF";
              else if (exp.dynAllocPlanets && (now<exp.startTime) && (exp.numPlanetsAllocated == 0))
                plcolor = "#FFEE00";
              else if (!exp.numPlanetsAllocated || (numPlanetsActive < exp.numPlanetsAllocated))
                plcolor = "#FF0000";
              else
                plcolor = "#00B000";

              append("<td bgcolor=\"%s\" align=\"right\">%d/%d</td>", plcolor, numPlanetsActive, exp.numPlanetsAllocated);

              const char *scolor = "#FFFFFF";
              if ((exp.startTime>now) || (exp.endTime<now))
                scolor = "#FFFFFF";
              else if (exp.dynAllocSats && exp.numSatellitesMissing)  
                scolor = "#FFEE00";
              else if (!numSatellitesActive)
                scolor = "#FF0000";
              else if (numSatellitesActive < exp.numSatellitesAllocated)
                scolor = "#FFEE00";
              else 
                scolor = "#00B000";

              if (exp.dynAllocSats && exp.numSatellitesMissing)
                append("<td bgcolor=\"%s\" align=\"right\">%d/%d (+%d)</td>", scolor, numSatellitesActive, exp.numSatellitesAllocated, exp.numSatellitesMissing);
              else
                append("<td bgcolor=\"%s\" align=\"right\">%d/%d</td>", scolor, numSatellitesActive, exp.numSatellitesAllocated);

              append("</tr>\n");
            }
          }
        }
        append("  </table><p>\n");
        
        /* List of planets */
        
        append("<h3>Planets</h3>\n");
        append("  <table border=\"1\" cellpadding=\"3\" cellspacing=\"0\"><tr><td><b>Address</b></td><td><b>Port</b></td><td><b>Status</b></td>%s</tr>\n", identifierFilter ? "" : "<td><b>Experiment</b></td>");
        for (int i=0; i<sun->numPlanetRecords(); i++) {
          int now = time(0);
          struct Sun::planetStruct planet;
          sun->statPlanetRecord(i, &planet);
          if ((!ownerFilter || !strcmp(planet.expOwner, ownerFilter)) && (!identifierFilter || !strcmp(planet.expIdentifier, identifierFilter))) {
            const char *stateName = "???", *stateColor = "#0000FF";
            if (planet.lastStatusReport < (now-Sun::PLANET_TIMEOUT_SEC)) {
              stateName = "DOWN";
              stateColor = "#FF0000";
            } else if (!sun->isPlanetConfiguredCorrectly(i)) {
              stateName = "CFG";
              stateColor = "#FFEE00";
            } else {
              stateName = "OK";
              stateColor = "#00B000";
            }

            append("  <tr><td>%s</td><td align=\"right\">%d</td><td bgcolor=\"%s\" align=\"center\">%s</td>", planet.address, planet.port, stateColor, stateName);
            if (!identifierFilter)
              append("<td><a href=\"/experiments/%s/%s\">%s%s%s</a></td>", planet.expOwner, planet.expIdentifier, ownerFilter ? "" : planet.expOwner, ownerFilter ? "" : ":", planet.expIdentifier);
            append("</tr>\n");
          }
        }

        append("</table><p>\n");

        /* List of satellites */

        append("<h3>Satellites</h3>\n");
        append("  <table border=\"1\" cellpadding=\"3\" cellspacing=\"0\"><tr><td><b>Identifier</b></td><td><b>Status</b></td><td><b>Domain</b></td>%s<td><b>Planet</b></td></tr>\n", identifierFilter ? "" : "<td><b>Experiment</b></td>");
        for (int i=0; i<sun->numSatelliteRecords(); i++) {
          int now = time(0);
          struct Sun::satelliteStruct sat;
          sun->statSatelliteRecord(i, &sat);

          if (((isAdmin && !ownerFilter && !identifierFilter) || sat.isAllocated) && (!ownerFilter || !strcmp(sat.expOwner, ownerFilter)) && (!identifierFilter || !strcmp(sat.expIdentifier, identifierFilter))) {
            const char *stateName = "???", *stateColor = "#0000FF";
            if (sat.lastStatusReport < (now-Sun::SATELLITE_TIMEOUT_SEC)) {
              stateName = "DOWN";
              stateColor = "#FF0000";
            } else if (sat.isAllocated && sat.isBootstrapping) {
              stateName = "BOOT";
              stateColor = "#FFEE00";
            } else {
              stateName = "OK";
              stateColor = "#00B000";
            }

            if (isAdmin) 
              append("  <tr><td><a href=\"/satellite/%s/%s\">%s %s</a></td>", sat.email, sat.identifier, sat.email, sat.identifier);
            else
              append("  <tr><td>%s</td>", renderBytes(sat.planetPassword, sat.planetPasswordLen, buf1));
              
            append("<td bgcolor=\"%s\" align=\"center\">%s</td><td>%s</td>", stateColor, stateName, sat.domain);
            if (!identifierFilter) {
              append("<td>");
              if (sat.isAllocated)
                append("<a href=\"/experiments/%s/%s\">%s%s%s</a>", sat.expOwner, sat.expIdentifier, ownerFilter ? "" : sat.expOwner, ownerFilter ? "" : ":", sat.expIdentifier);
              else
                append("-");
              append("</td>");
            }
            if (sat.planetName[0])
              append("<td>%s:%d</td></tr>\n", sat.planetName, sat.planetPort);
            else
              append("</td><td>-</td></tr>\n");
          }
        }
        append("</table><p>\n");
        
        if (identifierFilter) {
          append("<form method=\"get\" action=\"/reassociate/%s/%s\"><input type=\"submit\" name=\"action\" value=\"Reassociate satellites with planets\"></form>\n", ownerFilter, identifierFilter);
          append("<a href=\"/experiments/%s\">View all experiments owned by '%s'</a><p>", ownerFilter, ownerFilter);
        } else if (isAdmin && ownerFilter) {
          append("<a href=\"/experiments/\">View all experiments</a><p>");
        }
        
      } else {
        authenticationFailed(index, "SatelliteLab");
      }
    } else if (!strncmp(url, "/satellite/", 11)) {
    
      /* Satellite status page */
    
      if (!strcmp(auth, adminAuth)) {
        append("HTTP/1.1 200 OK\r\nServer: SatelliteLab\r\nContent-Type: text/html\r\nConnection: close\r\n\r\n");

        char *email = strtok(&url[11], "/");
        char *identifier = email ? strtok(NULL, "/\r\n") : NULL;
        append("<html><head><title>Satellite: %s %s</title></head>\n<body>\n", email, identifier);
        int idx = -1;
        for (int i=0; i<sun->numSatelliteRecords(); i++) {
          struct Sun::satelliteStruct sat;
          if (sun->statSatelliteRecord(i, &sat)) {
            if (!strcmp(sat.email, email) && !strcmp(sat.identifier, identifier))
              idx = i;
          }
        }

        if (idx >= 0) {
          struct Sun::satelliteStruct sat;
          sun->statSatelliteRecord(idx, &sat);
          append("<h3>Satellite: %s %s</h3>\n", email, identifier);
          append("<table border=\"1\">\n");
          append("<tr><td><b>Email:</b></td><td>%s</td></tr>\n", sat.email);
          append("<tr><td><b>Identifier:</b></td><td>%s</td></tr>\n", sat.identifier);
          append("<tr><td><b>Domain:</b></td><td>%s</td></tr>\n", sat.domain);
          append("<tr><td><b>Address:</b></td><td>%s:%d</td></tr>\n", ipdisp(sat.ip, buf1), sat.port);
          const char *os[] = { "Unknown", "Unix", "Java", "MacOS", "Windows" };
          int osCode = (sat.version>>8)&0xFF;
          append("<tr><td><b>Software:</b></td><td>%04X (%s)</td></tr>\n", sat.version, ((osCode>=0) && (osCode<5)) ? os[osCode] : "Unknown");
          append("<tr><td><b>Sun password:</b></td><td>%s</td></tr>\n", renderBytes(sat.sunPassword, 8, buf1));
          time_t ti = sat.lastWentOnline;
          append("<tr><td><b>Tracking since:</b></td><td>%s</td></tr>\n", ctime(&ti));
          ti = sat.lastStatusReport;
          append("<tr><td><b>Last report:</b></td><td>%s (%d seconds ago)</td></tr>\n", ctime(&ti), time(0)-(sat.lastStatusReport));
          ti = sat.nextRefreshAllocation;
          append("<tr><td><b>Next refresh:</b></td><td>%s</td></tr>\n", ctime(&ti));
          append("<tr><td><b>Bytes sent:</b></td><td>%u</td></tr>\n", sat.totalBytesSent);
          append("<tr><td><b>Bytes received:</b></td><td>%u</td></tr>\n", sat.totalBytesReceived);
          append("<tr><td><b>Transmit counter:</b></td><td>%u</td></tr>\n", sat.reportedBytesSent);
          append("<tr><td><b>Receive counter:</b></td><td>%u</td></tr>\n", sat.reportedBytesReceived);
          if (sat.isAllocated) {
            append("<tr><td><b>Experiment:</b></td><td>%s:%s</td></tr>\n", sat.expOwner, sat.expIdentifier);
            append("<tr><td><b>Planet password:</b></td><td>%s</td></tr>\n", renderBytes(sat.planetPassword, sat.planetPasswordLen, buf1));
            append("<tr><td><b>Planet:</b></td><td>%s (%s:%d)</td></tr>\n", sat.planetName, ipdisp(sat.planetIP, buf1), sat.planetPort);
          } else {
            append("<tr><td><b>Experiment:</b></td><td>-</td></tr>\n");
          }

          if (sat.isBootstrapping) {
            append("<tr><td><b>Bootstrapping:</b></td><td>yes</td></tr>\n");
            append("<tr><td><b>Bootstrap ID:</b></td><td>%08X</td></tr>\n", sat.bootstrapID);
            append("<tr><td><b>Last changed:</b></td><td>%d second(s) ago</td></tr>\n", time(0)-sat.lastChangedBootstrap);
            append("<tr><td valign=\"top\"><b>Bootstraps:</b></td><td>");
            for (int i=0; i<sat.numBootstraps; i++) {
              if (sat.currentBootstrapIndex == i)
                append("<b>");
              append("%s (%s:%d) %d/%d ", sat.bootAddress[i], ipdisp(sat.bootIP[i], buf1), sat.bootPort[i], sat.bootSuccesses[i], sat.bootAttempts[i]);
              if (sat.bootBestRTT[i] < INFINITE_RTT)
                append("%dms", sat.bootBestRTT[i]);
              else
                append("-");
              if (sat.currentBootstrapIndex == i)
                append("</b>");
              append("<br>\n");
            }
            append("</td></tr>\n");
          } else {
            append("<tr><td><b>Bootstrapping:</b></td><td>no</td></tr>\n");
          }
          append("</table><p>\n");
        } else {
          append("At this time, satellite <b>%s %s</b> is neither allocated to an experiment nor connected to this sun.<p>", email, identifier);
        }

        append("<a href=\"/\">Back to overview</a>\n");
      } else {
        authenticationFailed(index, "SatelliteLab");
      }
    } else {
      append("HTTP/1.1 404 NOT FOUND\r\nServer: SatelliteLab\r\nConnection: close\r\n\r\nNot found: %s\n", url);
    }

    if (!strncmp(&connection[index].buf[13], "OK", 2))
      append("<p><hr>Sun %s at %s (built on %s)\n</body></html>", VERSION, hostname, __DATE__);
  } else {
    strcpy(connection[index].buf, "HTTP/1.1 500 Internal server error\r\nServer: SatelliteLab\r\nConnection: close\r\n\r\n");
  }
  
  connection[index].ptr = 0;
  connection[index].isReading = false;
  selector->setDescriptorCallback(connection[index].fd, false, true, false, this);
}

void WebFrontend::descriptorReady(int fd, bool canRead, bool canWrite, bool hasException)
{
  /* Accept new connections */

  if (fd == httpSocket) {
    struct sockaddr_in from;
    socklen_t fromlen = sizeof(from);
    int fd = accept(httpSocket, (struct sockaddr*)&from, &fromlen);
    if (fd >= 0) {
      int idx = -1;
      for (int i=0; (idx<0) && (i<MAX_CONNECTIONS); i++) {
        if (!connection[i].active)
          idx = i;
      }

      if (idx >= 0) {
        connection[idx].active = true;
        connection[idx].fd = fd;
        connection[idx].isReading = true;
        connection[idx].ptr = 0;
        connection[idx].len = 0;
        selector->setDescriptorCallback(fd, true, false, false, this);
      } else {
        close(fd);
      }
    }
    
    return;
  }

  /* HTTP reading and writing */

  for (int i=0; i<MAX_CONNECTIONS; i++) {
    if (connection[i].active && (connection[i].fd == fd)) {
      if (connection[i].isReading) {
        assert(canRead);
        int r = read(connection[i].fd, &connection[i].buf[connection[i].len], sizeof(connection[i].buf) - connection[i].len);
        if (r > 0) {
          log->write(4, "Connection #%d read %d bytes", i, r);
          connection[i].len += r;
          maybeHandleHttpRequest(i);
        } else {
          log->write(4, "Connection #%d closed (read close)", i);
          close(connection[i].fd);
          selector->cancelDescriptorCallback(fd, this);
          connection[i].active = false;
        }
      } else {
        assert(canWrite);
        int w = write(connection[i].fd, &connection[i].buf[connection[i].ptr], connection[i].len - connection[i].ptr);
        if (w > 0) {
          log->write(4, "Connection #%d wrote %d bytes", i, w);
          connection[i].ptr += w;
          if (connection[i].ptr >= connection[i].len) {
            log->write(4, "Connection #%d closed (finished)", i);
            close(connection[i].fd);
            selector->cancelDescriptorCallback(fd, this);
            connection[i].active = false;
          }
        } else {
          log->write(4, "Connection #%d closed (write close)", i);
          close(connection[i].fd);
          selector->cancelDescriptorCallback(fd, this);
          connection[i].active = false;
        }
      }
      
      return;
    }
  }
  
  assert(false);
}
